#! /usr/bin/env python
#
#     Copyright (C) 2005 Ronan Keegan
#
#     This code is distributed under the terms and conditions of the
#     CCP4 Program Suite Licence Agreement as a CCP4 Application.
#     A copy of the CCP4 licence can be obtained by writing to the
#     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
#
# A MrBUMP wrapper for Refmac 
# Ronan Keegan - 26/1/07
#

import os, sys
import time

if not os.environ.has_key('CCP4'):
    raise RuntimeError, 'CCP4 not found'
if not os.environ.has_key('CBIN'):
    raise RuntimeError, 'CBIN not found'

class RefmacEXE:
   """ A class to run a Refmac job. """

   def __init__(self):
      if os.name=="nt":
         self.refmacEXE = os.path.join(os.environ["CBIN"], "refmac5.exe")  
      else:
         self.refmacEXE = os.path.join(os.environ["CBIN"], "refmac5")  

      self.logfile = ""
      self.hklin = ""
      self.hklout = ""
      self.pdbin = ""
      self.pdbout = ""
      self.key = ""
      self.keywords = []
      self.termination=False

      try:
         self.debug=eval(os.environ['MRBUMP_DEBUG'])
      except:
         self.debug=False

      self.local_debug = False

   def set_logfile(self, filename):
      self.logfile=filename

   def set_hklin(self, filename):
      self.hklin=filename

   def set_hklout(self, filename):
      self.hklout=filename

   def set_pdbin(self, filename):
      self.pdbin=filename

   def set_pdbout(self, filename):
      self.pdbout=filename

   def add_keyword(self, keyword):
      self.keywords.append(keyword)
      

   def run(self, hklin, hklout, pdbin, pdbout, logfile, keyfile=""):
      """ Function to launch a refmac job. """

      # Give a header output
      sys.stdout.write("#######################################\n")
      sys.stdout.write("Running Refmac...... \n")
      sys.stdout.write("#######################################\n")
      sys.stdout.write("\n")

      # Create a job and first set up the Reindex details

      # Check to see if the logfile has been specified, otherwise give it a default name
      if self.logfile=="":
         self.logfile=logfile 

      # Set HKLIN & HKLOUT if it has not been set already
      if self.hklin=="":
         self.hklin=hklin

      if self.hklout=="":
         self.hklout=hklout

      # Set PDBIN & PDBOUT 
      if self.pdbin=="":
         self.pdbin=pdbin

      if self.pdbout=="":
         self.pdbout=pdbout

      # Set the keywords for the job 
      if keyfile == "":
      # If no keyword file is given read keywords from the input keywords
         for keyword in self.keywords:
            self.key=self.key + keyword + "\n"

      else:
      # Read keywords from the keyword file 
         kfile=open(keyfile, "r")
         line=kfile.readline()
         while line:
            self.key = self.key + line
            line=kfile.readline()
         kfile.close()
        
      # Output the keywords if in debug mode
      if self.local_debug:
         sys.stdout.write("#########################\n")
         sys.stdout.write("Refmac keyword input:\n")
         sys.stdout.write("#########################\n")
         sys.stdout.write("\n")
         sys.stdout.write(self.key)
         sys.stdout.write("\n")
         time.sleep(2)
         
      # Set the command line
      command_line=self.refmacEXE + " HKLIN %s HKLOUT %s XYZIN %s XYZOUT %s" % (self.hklin, self.hklout, self.pdbin, self.pdbout)

      #Echo the command line in debug mode
      if self.local_debug:
        sys.stdout.write(command_line + "\n")
        sys.stdout.write("\n")

      # Open a pipe to the job
      i,oe=os.popen4(command_line)

      # Write the keywords to the job
      i.write(self.key)
      i.close()

      # Open the log file for writing
      log=open(self.logfile, "w")

      # Watch the output for successful termination
      out=oe.readline()
      log.write(out)

      while out:
         if self.local_debug:
            sys.stdout.write(out)

         if 'End of Refmac' in out:
            self.termination=True
         out=oe.readline()
         log.write(out)

      oe.close()
      log.close()

      # Check the output
      if self.termination == False: 
         sys.stdout.write("Refinement log: refmac failed to complete successfully\n")
         sys.stdout.write("\n")
         sys.stdout.write("Log file can be found here:\n")
         sys.stdout.write("   " + self.logfile + "\n")
         sys.stdout.write("\n")
      else:
         sys.stdout.write("Refinement log: Refmac completed successfully\n")
         sys.stdout.write("\n")
         sys.stdout.write("Output MTZ file written to:\n   " + self.hklout + "\n")
         sys.stdout.write("Output PDB file written to:\n   " + self.pdbout + "\n")
         sys.stdout.write("\n")

