########################################################################
# ccp4ipy_tests
#
# Python module providing test functions for the CCP4i module ccp4i.py.
#
# CVS_Id: $Id: ccp4ipy_tests.py,v 1.4 2008/04/02 15:28:46 rmk65 Exp $
#
#########################################################################

import ccp4i

########################################################################
# Test methods
#########################################################################

def test_dirlock():
    """Test if directories class can deal correctly with file locks."""
    
    print "test_dirlock"
    # Create a directories object
    dirs = ccp4i.directories()
    # Get the fullname of the directories.def file
    dirfile = dirs.source()
    print "\tSource file: \""+str(dirfile)+"\""
    # Check for a lock
    if dirs.islocked() == True:
        print "\tThe file is locked: test aborted"
        # Release the directories.def file
        dirs.release()
        return False
    print "\tThe file is unlocked"
    # Create a dummy lock file
    lock = ccp4i.lockfile(dirs.source())
    lock.lock()
    # Second attempt to open the file
    print "\tDummy lock created"
    dirs = ccp4i.directories()
    if dirs.islocked() == False:
        # The directories object has failed to detect the
        # dummy lock, this is an error
        print "\tThe lock is not recognised: test failed"
        dirs.release()
        return False
    print "\tThe lock is recognised: ok"
    # Attempt to open normally
    if dirs.load() == True:
        # The directories object should not load the data
        # if there is already a lock
        print "\tFile open without breaking lock: test failed"
        dirs.release()
        lock.unlock()
        return False
    print "\tFile failed to open: ok"
    # Attempt to open with force
    if dirs.load(force=True) == False:
        # The directories object should always return success
        # if the force option is specified
        print "\tFile open failed, lock wasn't forced: test failed"
        dirs.release()
        lock.unlock()
        return False
    print "\tFile open forced: ok"
    # Verify that the data is loaded
    print "\t***Verifying that directories data is readable***"
    if dirs.isloaded() == True:
        try:
            print "\tObject believes it is loaded"
            print "\tN_PROJECTS = "+str(dirs.n_projects())
            print "\tProjects = "+str(dirs.listprojects())
            print "\tDef dirs = "+str(dirs.listdefdirs())
            i = 0
            for proj in dirs.listprojects():
                i=i+1
                print "\n\t===================="
                print "\tCheck project #"+str(i)
                print "\t===================="
                if test_openproject(proj,dirs.projectdir(proj)) == False:
                    # There was a problem checking the
                    # project
                    dirs.release()
                    return False
        except:
            # One of the operations above raised an exception
            print "\t***Attempting to access some attributes failed***"
            print "\tTest failed"
            dirs.release()
            return False
    # Finish
    dirs.release()
    return True

def test_openproject(project,projectdir):
    """Try to open and verify a CCP4i project."""

    print "\tProject: "+str(project)+"  Dir: "+str(projectdir)
    # Initialise
    db = ccp4i.database(project,projectdir)
    if db.islocked():
        print "\t***Warning: project db is already locked"
        db.close()
        return False
    # Open the project
    if db.open() == False:
        print "\tFailed to open the project database"
        db.close()
        return False
    # Report some data
    print "\t\tNJOBS = "+str(db.njobs())
    print "\t\tJobs  = "+str(db.listjobs())
    # Check that all jobs exist
    for job in db.listjobs():
        if not test_jobexists(db,job):
            # Job wasn't verified
            print "\t***Job #"+str(job)+" not verified***"
            db.close()
            return False
    print "\tAll jobs verified"
    # Finished
    db.close()
    return True

#######################################################################
# Supporting tests and utility functions
#######################################################################

def test_jobexists(db,job):
    """Verify that the specified job exists in the database."""

    # Set of items required
    items = ["DATE","STATUS","TITLE","TASKNAME"]
    for item in items:
        if not db.itemexists(job,item):
            # An item is missing
            return False
    # All items found
    return True

def clear_all_dblocks():
    """Clear the locks on all project directories."""
    
    dirs = ccp4i.directories()
    dirs.load()
    try:
        for proj in dirs.listprojects():
            print "\tProject #"+str(proj)
            clear_dblock(str(proj),dirs.projectdir(proj))
            print "\t...done"
    except:
        print "Exception encountered: error clearing lock for projects"
        dirs.release()
        return False
    dirs.release()
    return True

def clear_dblock(project,projectdir):
    """Clear any lock on a project by (forced) open and close."""

    # Make a project object
    db = ccp4i.database(project,projectdir)
    if db.islocked():
        print "Project \""+str(project)+"\" is locked"
        db.open(grablock=True)
    else:
        print "Project \""+str(project)+"\" is not locked"
        db.open()
    db.close()
    return True

def make_test_project(name,dir):
    """Make a small test project and database.

    Specify the name of the project and the target directory (which must
    not already exist."""

    dirpath = ccp4i.GetAbsolutePath(dir)
    if ccp4i.DirExists(dirpath):
        print "Directory "+str(dirpath)+" already exists"
        return False
    db = ccp4i.database(name,dirpath)
    if not db.create():
        print "Failed to create new test project "+str(name)
        return False
    # Populate the database with dummy jobs
    db.newjob(taskname="smartie",status="FINISHED",title="A run of smartie")
    db.newjob(taskname="scala",status="FINISHED",title="This one worked")
    db.newjob(taskname="truncate",status="FAILED",title="This one didn't")
    return True

def announce_test_start(name):
    """Print a banner announcing the start of a test."""

    print "\n*********************************************"
    print "* "+str(name)+": STARTED"
    print "*********************************************"

def announce_test_end(name,result):
    """Print a banner announcing the results of a test."""

    print "*********************************************"
    print "* "+str(name)+": "+str(result)
    print "*********************************************\n"

#######################################################################
# Main method
#######################################################################

if __name__ == "__main__":
    """Run the tests."""

    announce_test_start("test_dirlock")
    if test_dirlock():
        result = "SUCCEEDED"
    else:
        result = "FAILED"
    announce_test_end("test_dirlock",result)
