#! /usr/bin/env python
#
#     Copyright (C) 2005 Ronan Keegan 
#
#     This code is distributed under the terms and conditions of the
#     CCP4 Program Suite Licence Agreement as a CCP4 Application.
#     A copy of the CCP4 licence can be obtained by writing to the
#     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
#
#
# Code to handle the keyword input to MRBUMP
# Ronan Keegan 12.10.05

import string
import sys
import os
import mol_weight

class Keywords:
   """ A class to read the keywords for MRBUMP from the stdin or input from a file. """

   def __init__(self):
      self.col_labels=dict([])
      self.JOBID=''
      self.parent_JOBID=''

      self.ROOTDIR = os.getcwd() 
      self.PDBDIR = ""
      self.PDBLOCAL = None
   
      self.NMASU=None
      self.PACK=5
      self.NCYC=30
      self.REFTWIN=False
      self.PIRCYC=3

      self.MRNUM = 10
      self.ENSEMNUM = 5
      self.USEENSEM = True
 
      self.E_value = 0.02

      self.MR_PROGRAM_LIST = ['MOLREP']
      self.MAPROGRAM = 'MAFFT'
      
      self.SSM = False
      self.SCOP = True 
      self.PQS = True 
      self.PISA = False

      self.MDLDPDBCLP = False 
      self.MDLPLYALA = False 
      self.MDLMOLREP = True 
      self.MDLCHAINSAW = True 

      self.DEBUG = False
      self.CLUSTER = False
      self.QTYPE = "SGE"
      self.PROXYSERVER = ""

      self.fasta_local = True
      self.run_local   = False

      self.UPDATE = True
      self.DOFASTA = True
      self.TRYALL = False
      self.CLEAN = False
      self.CHECK = True

      self.HTMLOUT = False
      self.XMLOUT = False
      self.XMLFILE = ""
      self.results_html =  ''

      self.ignore_list = []
      self.include_list = []
      self.local_list = dict([])
      self.locfile_count = 0
      self.error_string = ''

      self.labin_found = False
      self.jobid_found = False

      self.USEACORN = False
      self.ACORNRES = 1.7

      self.USECPIRATE = False

      self.ENANT = False
      self.DUMMY = False
      self.PICKLE = True

      self.DBOUT = False
      self.DBVIEW = False

      self.fixed_list = dict([])
      self.fixedfile_count = 0
      self.FIXED = False

   def setErrorString(self, error_string):
      self.error_string = error_string

   def read_keywords(self):
      """ A function to read in the keyword arguments from the STDIN or from a file. """

      print "ENTER KEYWORD INPUT FROM FILE OR FROM STANDARD INPUT"

      while True:
         data_line = string.strip(raw_input("-->\n"))
         if data_line == '':
           keyword = "NULL"
         else:
           keyword = string.split(data_line)[0]

         if "LABI" in keyword[0:4].upper():
            self.parse_LABIN(data_line)
            self.labin_found = True
    
         elif "JOBI" in keyword[0:4].upper():
            self.parse_JOBID(data_line)
            self.jobid_found = True
            
         elif "ROOT" in keyword[0:4].upper():
            self.parse_ROOTDIR(data_line)

         elif "PDBD" in keyword[0:4].upper():
            self.parse_PDBDIR(data_line)

         elif "PDBL" in keyword[0:4].upper():
            self.parse_PDBLOCAL(data_line)

         elif "IGNO" in keyword[0:4].upper():
            self.parse_IGNORE(data_line)

         elif "INCL" in keyword[0:4].upper():
            self.parse_INCLUDE(data_line)

         elif "LOCA" in keyword[0:4].upper():
            self.parse_LOCALFILE(data_line)

         elif "NMAS" in keyword[0:4].upper():
            self.parse_NMASU(data_line)

         elif "PACK" in keyword[0:4].upper():
            self.parse_PACK(data_line)

         elif "NCYC" in keyword[0:4].upper():
            self.parse_NCYC(data_line)

         elif "PIRC" in keyword[0:4].upper():
            self.parse_PIRCYC(data_line)

         elif "MRNU" in keyword[0:4].upper():
            self.parse_MRNUM(data_line)

         elif "ENSE" in keyword[0:4].upper():
            self.parse_ENSEMNUM(data_line)

         elif "EVAL" in keyword[0:4].upper():
            self.parse_EVALUE(data_line)

         elif "MRPR" in keyword[0:4].upper():
            self.parse_MRPROGRAMS(data_line)

         elif "MAPR" in keyword[0:4].upper():
            self.parse_MAPROGRAM(data_line)

         elif "QTYP" in keyword[0:4].upper():
            self.parse_QTYPE(data_line)

         elif "PROX" in keyword[0:4].upper():
            self.parse_PROXYSERVER(data_line)

         elif "XMLF" in keyword[0:4].upper():
            self.parse_XMLFILE(data_line)

         elif "MDLD" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'MDLDPDBCLP')

         elif "MDLP" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'MDLPLYALA')

         elif "MDLM" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'MDLMOLREP')

         elif "MDLC" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'MDLCHAINSAW')

         elif "DEBU" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'DEBUG')

         elif "CLUS" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'CLUSTER')

         elif "SSMS" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'SSMSEARCH')

         elif "SCOP" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'SCOPSEARCH')

         elif "PQSS" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'PQSSEARCH')

         elif "PISA" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'PISASEARCH')

         elif "PWAL" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'PWALIGN')

         elif "FAST" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'FASTALOCAL')

         elif "RUNL" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'RUNLOCAL')
      
         elif "UPDA" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'UPDATE')
      
         elif "DOFA" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'DOFASTA')
      
         elif "TRYA" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'TRYALL')
      
         elif "CLEA" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'CLEAN')
      
         elif "HTML" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'HTMLOUT')
      
         elif "XMLO" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'XMLOUT')
      
         elif "RESH" in keyword[0:4].upper():
            self.parse_RESHTML(data_line)

         elif "USEA" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'USEACORN')
      
         elif "ACOR" in keyword[0:4].upper():
            self.parse_ACORNRES(data_line)

         elif "USEC" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'USECPIRATE')
      
         elif "USEE" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'USEENSEM')
      
         elif "CHEC" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'CHECK')
      
         elif "ENAN" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'ENANT')
      
         elif "DUMM" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'DUMMY')
      
         elif "PICK" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'PICKLE')
      
         elif "DBOU" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'DBOUT')
      
         elif "DBVI" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'DBVIEW')
      
         elif "REFT" in keyword[0:4].upper():
            self.parse_FLAG(data_line, 'REFTWIN')

         elif "FIXE" in keyword[0:4].upper():
            self.parse_FIXED_XYZIN(data_line)

         elif "NULL" in keyword.upper():
            print "Please enter a keyword or type END to finish keyword entry"
      
         elif "END" in keyword.upper():
            break
       
         else:
            print "Ignoring unrecognised keyword: %s" % keyword 
            
      # Verify the 'NUM_' keywords are correct
      self.verify_NUM()

      # Check that chains have been specified if DOFASTA set to False
      if self.DOFASTA == False and self.include_list == [] and len(self.local_list.keys()) == 0:
         self.keyword_usage('DOFASTA', \
            'If DOFASTA is set to false you need to specify chains ' \
          + 'to be used as template models using the INCLUDE or LOCALFILE keywords' , 19)

      # Check that the JOBID keyword hs abeen provided. All others are optional.
      if self.jobid_found == False:
         self.keyword_usage('jobid', 'JOBID keyword not provided', 12)

   def parse_JOBID(self, data_line):
      """ A function to parse the JOBID keyword. """
      
      if len(string.split(data_line)) == 1:
         print "Keyword entry error: JOBID requires an argument"
      else:
         self.JOBID=string.split(data_line)[1]
      
         print "Job identifier: %s" % self.JOBID

   def parse_ROOTDIR(self, data_line):
      """ A function to parse the ROOT keyword. """
   
      if len(string.split(data_line)) == 1:
         print "Keyword entry error: ROOTDIR requires an argument"
      else:
         self.ROOTDIR=string.join(string.split(data_line)[1:], " ")
   
         # Test directory existence
         if os.path.isdir(self.ROOTDIR) == False:
            self.keyword_usage('root_dir', '%s does not exist' % self.ROOTDIR, 15)
      
         print "Root directory: %s" % self.ROOTDIR

   def parse_PDBDIR(self, data_line):
      """ A function to parse the keyword PDBDIR. PDBDIR is a directory specified by
          the user that holds PDB files for use in MRBUMP. This can help reduce the
          need for downloads from the web. PDB files should take the format:
             <PDB ID>.pdb   (e.g. 1nio.pdb) """

      if len(string.split(data_line)) == 1:
         print "Keyword entry error: PDBDIR requires an argument"
      else:
         self.PDBDIR=string.join(string.split(data_line)[1:], " ")
        
         # Test directory existence
         if os.path.isdir(self.PDBDIR) == False:
            sys.stdout.write("Keyword Warning: The PDBDIR was not found:\n  %s\n" % self.PDBDIR)
         else:
            sys.stdout.write("Input PDB files directory: %s\n" % self.PDBDIR)
         sys.stdout.write("\n")

   def parse_PDBLOCAL(self, data_line):
      """ A function to parse the keyword PDBLOCAL. PDBLOCAL specifies a local mirror of the
          PDB. The mirror must have the files stored under directories specified by the 
          middle two characters in the PDB id. The files must also be zipped and of the 
          format pdb1nio.ent.gz """

      if len(string.split(data_line)) == 1:
         print "Keyword entry error: PDBLOCAL requires an argument (directory path)"
      else:
         self.PDBLOCAL=string.join(string.split(data_line)[1:], " ")
        
         # Test directory existence
         if os.path.isdir(self.PDBLOCAL) == False:
            sys.stdout.write("Keyword Error: PDBLOCAL directory was not found:\n  %s\n" % self.PDBLOCAL)
            sys.stdout.write("               This should be the path to the local PDB Mirror\n")
            sys.stdout.write("               Remove this keyword if you do not have a local PDB Mirror\n")
            sys.stdout.write("\n")
            sys.exit(1)
         # Do a couple of tests to see if it is in the correct format
         elif os.path.isdir(os.path.join(self.PDBLOCAL, "ni")) == False:
            sys.stdout.write("Keyword Error: PDBLOCAL directory does not have the standard PDB Mirror format\n")
            sys.stdout.write("               It should have files stored under subdirectories denoted by the\n")
            sys.stdout.write("               middle two characters of the PDB codes stored therin\n")
            sys.stdout.write("               e.g. file pdb1nio.ent.gz should be stored in directory 'ni' (dir)\n")
            sys.stdout.write("\n")
            sys.exit(1)
         elif os.path.isfile(os.path.join(self.PDBLOCAL, "ni", "pdb1nio.ent.gz")) == False:
            sys.stdout.write("Keyword Error: PDBLOCAL directory does not have the standard PDB Mirror format\n")
            sys.stdout.write("               It should have files stored under subdirectories denoted by the\n")
            sys.stdout.write("               middle two characters of the PDB codes stored therin\n")
            sys.stdout.write("               e.g. file pdb1nio.ent.gz should be stored in directory 'ni' (file)\n")
            sys.stdout.write("\n")
            sys.exit(1)
         else:
            sys.stdout.write("Local PDB Mirror directory: %s\n" % self.PDBLOCAL)
         sys.stdout.write("\n")

   def parse_RESHTML(self, data_line):
      """ A function to parse the RESHTML keyword. """
   
      if len(string.split(data_line)) == 1:
         print "Keyword entry error: RESHTML requires an argument"
      else:
         self.results_html=string.split(data_line)[1]
   
         print "If Results in HTML required the file will be written to: %s" % self.results_html

   def parse_NMASU(self, data_line):
      """ A function to parse the NMASU keyword. """

      if len(string.split(data_line)) == 1:
         print "No value given for NMASU. Number of molecules will be calculated automatically"
      else:
         value = string.split(data_line)[1]

         if value.isdigit():
            self.NMASU=int(value)
         else:
            self.keyword_usage('NMASU', "'%s' is not a valid value for NMASU, must be integer" % value, 4)

         print "Number of molecules in the asu: %d" % self.NMASU

   def parse_PACK(self, data_line):
      """ A function to parse the PACK keyword. """

      if len(string.split(data_line)) == 1:
         print "Keyword entry error: PACK requires an argument"
      else:
         value = string.split(data_line)[1]

         if value.isdigit():
            self.PACK=int(value)
         else:
            self.keyword_usage('PACK', "'%s' is not a valid value for PACK, must be integer" % value, 4)

         print "Number of clashes tolerated in Phaser: %d" % self.PACK

   def parse_NCYC(self, data_line):
      """ A function to parse the NCYC keyword. """

      if len(string.split(data_line)) == 1:
         print "Keyword entry error: NCYC requires an argument"
      else:
         value = string.split(data_line)[1]

         if value.isdigit():
            self.NCYC=int(value)
         else:
            self.keyword_usage('NCYC', "'%s' is not a valid value for NCYC, must be integer" % value, 4)

         print "Number of cycles of restrained refinement to use in Refmac = %d" % self.NCYC

   def parse_PIRCYC(self, data_line):
      """ A function to parse the PIRCYC keyword. """

      if len(string.split(data_line)) == 1:
         print "Keyword entry error: PIRCYC requires an argument"
      else:
         value = string.split(data_line)[1]

         if value.isdigit():
            self.PIRCYC=int(value)
         else:
            self.keyword_usage('PIRCYC', "'%s' is not a valid value for PIRCYC, must be integer" % value, 4)

         print "Number of cycles of phase improvement in Cpirate = %d" % self.PIRCYC

   def parse_MRNUM(self, data_line):
      """ A function to parse the MRNUM keyword. """
   
      if len(string.split(data_line)) == 1:
         print "Keyword entry error: MRNUM requires an argument"
      else:
         value = string.split(data_line)[1]

         if value.isdigit():
            self.MRNUM=int(value)
         else:
            self.keyword_usage('MRNUM', "'%s' is not a valid value for MRNUM, must be integer" % value, 5)
      
         print "Number of search models to use in MR: %d" % self.MRNUM

   def parse_ENSEMNUM(self, data_line):
      """ A function to parse the ENSEMNUM keyword. """
   
      if len(string.split(data_line)) == 1:
         print "Keyword entry error: ENSEMNUM requires an argument"
      else:
         value = string.split(data_line)[1]

         if value.isdigit():
            self.ENSEMNUM=int(value)
         else:
            self.keyword_usage('ENSEMNUM', "'%s' is not a valid value for ENSEMNUM, must be integer" % value, 6)
      
         print "Number of models to use in Phaser ensemble: %d" % self.ENSEMNUM

   def parse_EVALUE(self, data_line):
      """ A function to parse the E_value keyword. """

      if len(string.split(data_line)) == 1:
         print "Keyword entry error: EVALUE requires an argument"
      else:
         value=string.split(data_line)[1]  
 
         try:
            self.E_value=float(value)
         except:
            self.keyword_usage('E_value', "'%s' is not a valid value for EVALUE, must be float" % value, 7)
      
         print "E value to use in fasta search: %.6f" % self.E_value

   def parse_MRPROGRAMS(self, data_line):
      """ A function to parse the MRPROGRAMS keyword. """

      self.MR_PROGRAM_LIST = []
 
      if len(string.split(data_line)) == 1:
         print "Keyword entry error: MRPROGRAMS requires one or more arguments"
         print "Setting default Molecular Replacement program to Molrep..."
         self.MR_PROGRAM_LIST = ['MOLREP']
      else:
         num_programs=len(string.split(data_line)) - 1
         for i in range(num_programs):
            program = (string.split(data_line)[i+1]).upper()
            if program != 'MOLREP' and program != 'PHASER':     
               self.keyword_usage('MRPROGRAMS', "'%s' is not a valid MR program" % program, 10)
            else:
               self.MR_PROGRAM_LIST.append(program)
      
         sys.stdout.write("MR programs to be used: ")
         for prog in self.MR_PROGRAM_LIST: 
            sys.stdout.write(prog +  " ")
         sys.stdout.write("\n")

   def parse_MAPROGRAM(self, data_line):
      """ A function to parse the MAPROGRAM keyword. """

      if len(string.split(data_line)) == 1:
         print "Keyword entry error: MAPROGRAM requires an argument"
      else:
         program = (string.split(data_line)[1]).upper()

         if program != 'MAFFT' and program != 'CLUSTALW' and program != 'CLUSTALW2' and program != 'PROBCONS' and program != 'T_COFFEE':     
            self.keyword_usage('MAPROGRAM', "'%s' is not a valid Sequence Multiple Alignment program" % program, 14)
         else:
            self.MAPROGRAM=program
      
         print "Default sequence alignment program: %s" % self.MAPROGRAM

   def parse_QTYPE(self, data_line):
      """ A function to parse the QTYPE keyword. """

      if len(string.split(data_line)) == 1:
         print "Keyword entry error: QTYPE requires an argument"
      else:
         qtype = (string.split(data_line)[1]).upper()

         if qtype != 'SGE' and qtype != 'PBS':     
            self.keyword_usage('QTYPE', "'%s' is not a recogniesed queuing system. 'SGE' or 'PBS' are accepted. " % qtype, 14)
         else:
            self.QTYPE=qtype
      
         print "Specified queuing system on cluster: %s" % self.QTYPE

   def parse_PROXYSERVER(self, data_line):
      """ A function to parse the PROXYSERVER keyword. """

      if len(string.split(data_line)) == 1:
         print "Keyword entry error: PROXYSERVER requires an argument"
      else:
         self.PROXYSERVER = (string.split(data_line)[1]).lower()
         
         print "Proxy server has been set to: %s" % self.PROXYSERVER
         print "This will be set in the environment and will be used for making connections"
         print "to the various online services used by MrBUMP (e.g. for PDB downloads)"
         print ""

   def parse_IGNORE(self, data_line):
      """ A function to parse the IGNORE keyword. """
   
      # Make a list from the line and delete the keyword IGNORE from the start
      line=string.split(data_line)
      del line[0]
      
      # Verify each PDB ID and add it to the ignore list 
      for i in line:
         pdb_id=i.lower()
         if len(pdb_id) !=4 or pdb_id.isalnum() == False or pdb_id[0].isdigit() == False:
            self.keyword_usage('ignore', "%s is not a valid PDB ID" % pdb_id, 3)
         else:
            self.ignore_list.append(pdb_id)
      print "List of PDB codes to ignore in the search: ", self.ignore_list
      
   def parse_INCLUDE(self, data_line):
      """ A function to parse the INCLUDE keyword. """
   
      # Make a list from the line and delete the keyword INCLUDE from the start
      line=string.split(data_line)
      del line[0]
      
      # Verify each PDB ID and add it to the include list 
      for i in line:
         chain=i.lower()
         if len(chain) !=6 or chain[0].isdigit() == False or chain[4] != "_" or chain[5].isalpha() == False:
            self.keyword_usage('include', "%s is not a valid Chain ID" % chain, 3)
         else:
            self.include_list.append(chain)
      print "List of PDB codes to include in the search: ", self.include_list
      
   def parse_ACORNRES(self, data_line):
      """ A function to parse the Acorn resolution limit keyword. """

      if len(string.split(data_line)) == 1:
         print "Keyword entry error: ACORNRES requires an argument"
      else:
         value=string.split(data_line)[1]  
 
         try:
            self.ACORNRES=float(value)
         except:
            self.keyword_usage('ACORNRES', "'%s' is not a valid value for ACORNRES, must be float" % value, 7)
      
         print "Acorn will be invoked if the collected data resolution is better than: %.3f Angstroms" % self.ACORNRES

   def parse_XMLFILE(self, data_line):
      """ A function to parse the XMLFILE name. """

      if len(string.split(data_line)) == 1:
         print "Keyword entry error: XMLFILE requires an argument (full path to a file name)"
      else:
         filename=string.split(data_line)[1]  

         # Check to see if the full path has been given, if not use the current directory
         dir_file=os.path.split(filename)

         if dir_file[0] == "":
            dir=os.getcwd()
            self.XMLFILE=os.path.join(dir, filename)
         else:
            self.XMLFILE=filename

         sys.stdout.write("XMLFILE: If XMLOUT is requested details of the job will be written to the XML file:\n   %s\n" % self.XMLFILE)
         sys.exit()

   def parse_LOCALFILE(self, data_line):
      """ A function to parse the LOCALFILE keyword. """
   
      if "LOCALFILE" in data_line.upper() and "CHAIN" in data_line.upper() and "RMS" in data_line.upper():
         # Pick out the filename (accoutning for spaces)
         filename=self.get_FILENAME(data_line, ["CHAIN", "RMS"])

         # Pick out the chain name
         line=string.split(data_line.upper())
         for i in range(len(line)):
            if line[i] == "CHAIN":
               if len(line) > i+1:
                  chainID=line[i+1] 
               else:
                  error_message="No argument given for CHAIN subkeyword for keyword LOCALFILE"
                  self.LOCALFILE_usage(error_message)
                  sys.exit()

         # Pick out the RMS value         
         line=string.split(data_line.upper())
         for i in range(len(line)):
            if line[i] == "RMS":
               if len(line) > i+1:
                  rms=line[i+1] 
               else:
                  error_message="No argument given for RMS subkeyword for keyword LOCALFILE"
                  self.LOCALFILE_usage(error_message)
                  sys.exit()

         if self.is_a_float(rms):
            rms=float(rms)
         else:
            error_message="Incorrect type for RMS in LOCALFILE keyword. Should be a float"
            self.LOCALFILE_usage(error_message)
            sys.exit()
            
      elif "LOCALFILE" in data_line.upper() and "CHAIN" in data_line.upper():
         # Pick out the filename (accoutning for spaces)
         filename=self.get_FILENAME(data_line, ["CHAIN"])
     
         # Pick out the chain name
         line=string.split(data_line.upper())
         for i in range(len(line)):
            if line[i] == "CHAIN":
               if len(line) > i+1:
                  chainID=line[i+1] 
               else:
                  error_message="No argument given for CHAIN subkeyword for keyword LOCALFILE"
                  self.LOCALFILE_usage(error_message)
                  sys.exit()

         # Set the rms value to be 0.0
         # If rms is set to 0.0 the sequence ID will be used for RMS input to phaser
         rms=0.0

      elif "LOCALFILE" in data_line.upper() and "RMS" in data_line.upper():
         # Pick out the filename (accoutning for spaces)
         filename=self.get_FILENAME(data_line, ["RMS"])
      
         # Pick out the RMS value         
         line=string.split(data_line.upper())
         for i in range(len(line)):
            if line[i] == "RMS":
               if len(line) > i+1:
                  rms=line[i+1] 
               else:
                  error_message="No argument given for RMS subkeyword for keyword LOCALFILE"
                  self.LOCALFILE_usage(error_message)
                  sys.exit()

         if self.is_a_float(rms):
            rms=float(rms)
         else:
            error_message="Incorrect type for RMS in LOCALFILE keyword. Should be a float"
            self.LOCALFILE_usage(error_message)
            sys.exit()
            
         # Set the chainID to be "A"
         chainID="A"

      elif "LOCALFILE" in data_line.upper():
         # Pick out the filename (accoutning for spaces)
         filename=self.get_FILENAME(data_line)
 
         # Set the chainID to be "A" and rms to be 0.0
         # If rms is set to 0.0 the sequence ID will be used for RMS input to phaser
         chainID="A"
         rms=0.0

      else:
         error_message="Incorrect arguments for LOCALFILE keyword"
         self.LOCALFILE_usage(error_message)
         sys.exit()

      # Verify that the file exists
      if os.path.isfile(filename) == False:
         error_message="File specified in LOCALFILE keyword not found:\n  %s\nDid you specify the full path?" % filename
         self.LOCALFILE_usage(error_message)
         sys.exit()
      else:
         # Set up a LocalPDB class for this local file
         lpdb=LocalPDB()
         lpdb.filename=filename
         lpdb.filestring=os.path.splitext(os.path.split(lpdb.filename)[1])[0]
         lpdb.chainID=chainID
         lpdb.rms=rms
         lpdb.number=self.locfile_count

         # Add this local file class to the local_list dictionary
         self.local_list["loc" + `self.locfile_count` + "_" + chainID + "_" + lpdb.filestring] = lpdb
         self.locfile_count = self.locfile_count + 1
      
      # Echo the details to the stdout
      sys.stdout.write("Chain from local file to include in model search:\n")
      if rms==0.0:
         sys.stdout.write("Chain ID: %s -- Filename: %s\n" % (chainID, filename)) 
      else:
         sys.stdout.write("Chain ID: %s (RMS=%.2f) -- Filename: %s\n" % (chainID, rms, filename)) 


   def LOCALFILE_usage(self, error_message):
      """ The usage message for the LOCALFILE keyword. """

      sys.stdout.write("Keyword input error: %s\n" % error_message)      
      sys.stdout.write("Format: LOCALFILE <filename> [CHAIN <chain ID>] [RMS <rms value>]\n")      
      sys.stdout.write("\n")      
     
   def parse_FIXED_XYZIN(self, data_line):
      """ A function to parse the FIXED_XYZIN keyword. """
   
      # The line format should be:
      #    FIXED_XYZIN <pdbfile> IDENTITY <?> 
      # This can occur multiple times

      line=string.split(data_line)
      if "FIXE" not in line[0].upper()  or "IDEN" not in data_line.upper():
         sys.stdout.write("Keyword input error: FIXED_XYZIN\n") 
         sys.stdout.write("  Format: FIXED_XYZIN <filename> IDENTIY <fractional sequence identity>\n") 
         sys.stdout.write("     e.g. FIXED_XYZIN /home/user/foo.pdb IDENTIY 0.6\n") 
         sys.stdout.write("\n") 
         sys.exit()

      # Set the identity value
      # Test to see if it's a valid float
      if self.is_a_float(line[-1]):
         identity=float(line[-1])
      else:
         sys.stdout.write("Keyword input error: non-float value for IDENTITY in line:\n")
         sys.stdout.write("   %s\n" % data_line)
         sys.stdout.write("\n")
         sys.exit()

      # Strip away everything except the file name
      file=line
      del file[-1]  # strips the identity value
      del file[-1]  # strips the "IDENTITY" keyword
      del file[0]   # strips the "FIXED_XYZIN" keyword

      # Catch for any spaces in the filename (e.g. if on windows the file is in Docs & Settings)
      filename=string.join(file)

      # Check to see if the file actually exists
      if os.path.isfile(filename) == False:
         sys.stdout.write("Keyword input error: File not found\n")
         sys.stdout.write("  File: %s was not found\n" % filename)
         sys.stdout.write("\n")
         sys.exit()

      # Work out the Molecular Weight for this component
      mw=mol_weight.MolWeight()
      molecular_weight=mw.get_pdbMW(filename)

      # Setup the fixed pdb file class for this file
      fixed_pdb=FixedPDB()
      fixed_pdb.filename=filename
      fixed_pdb.identity=identity
      fixed_pdb.mol_weight=molecular_weight

      # Add this file class to the fixed_list dictionary
      self.fixed_list[self.fixedfile_count]=fixed_pdb

      # Increment the fixed files count 
      self.fixedfile_count = self.fixedfile_count + 1

      # Set the FIXED flag to True
      self.FIXED=True

      # Detail the file inclusion to the stdout
      sys.stdout.write("Fixed component to be used in MR search added\n")
      sys.stdout.write("   Filename: %s\n" % filename)
      sys.stdout.write("   Sequence identity for this component: %.3f\n" % identity)
      sys.stdout.write("   Molecular weight for this component: %.3f\n" % molecular_weight)

   def parse_FLAG(self, data_line, env_flag):
      """ A function to parse any boolen flag keywords. """
  
      flag=string.split(data_line)[1]
       
      # Verify that the variable is set to True or False
      if flag.upper() != 'FALSE' and flag.upper() != 'TRUE' and flag != '1' and flag != '0':
         self.keyword_usage(env_flag, "'%s' is not a valid option for keyword %s" % (flag, env_flag), 15)

      if flag.upper() == 'TRUE' or flag == '1':
         if env_flag == 'DEBUG': 
            self.DEBUG=True
            os.environ["MRBUMP_DEBUG"]="True"
            print "Debug set to: ", self.DEBUG
         if env_flag == 'CLUSTER': 
            self.CLUSTER=True
            os.environ["MRBUMP_CLUSTER"]="True"
            print "Cluster set to: ", self.CLUSTER
         if env_flag == 'SCOPSEARCH':
            self.SCOP=True
            print "SCOP search set to: ", self.SCOP
         if env_flag == 'SSMSEARCH':
            self.SSM=True
            print "SSM search set to: ", self.SSM
         if env_flag == 'PQSSEARCH':
            self.PQS=True
            self.PISA=False
            #print "PQS search set to: ", self.PQS, "(and PISA search set to: ", self.PISA, " )"
            print "PQS search set to: ", self.PQS
         if env_flag == 'PISASEARCH':
            self.PISA=True
            self.PQS=False
            print "PISA search set to: ", self.PISA, "(and PQS search set to: ", self.PQS, " )"
         if env_flag == 'FASTALOCAL':
            self.fasta_local=True
            print "Run local Fasta search set to: ", self.fasta_local
         if env_flag == 'RUNLOCAL':
            self.run_local=True
            print "Run local set to: ", self.run_local
         if env_flag == 'UPDATE':
            self.UPDATE=True
            print "Update database files set to: ", self.UPDATE
         if env_flag == 'DOFASTA':
            self.DOFASTA=True
            print "Do the FASTA search set to: ", self.DOFASTA
         if env_flag == 'TRYALL':
            self.TRYALL=True
            print "Run in TRYALL mode set to: ", self.TRYALL
         if env_flag == 'CLEAN':
            self.CLEAN=True
            print "CLEAN flag set to: ", self.CLEAN
         if env_flag == 'HTMLOUT':
            self.HTMLOUT=True
            print "Run in HTMLOUT mode set to: ", self.HTMLOUT
         if env_flag == 'XMLOUT':
            self.XMLOUT=True
            print "Run in XMLOUT mode set to: ", self.XMLOUT
         if env_flag == 'MDLDPDBCLP':
            self.MDLDPDBCLP=True
            print "Generate PDBClip models set to: ", self.MDLDPDBCLP
         if env_flag == 'MDLPLYALA':
            self.MDLPLYALA=True
            print "Generate Polyalanine models set to: ", self.MDLPLYALA
         if env_flag == 'MDLMOLREP':
            self.MDLMOLREP=True
            print "Generate Molrep models set to: ", self.MDLMOLREP
         if env_flag == 'MDLCHAINSAW':
            self.MDLCHAINSAW=True
            print "Generate Chainsaw models set to: ", self.MDLCHAINSAW
         if env_flag == 'USEACORN':
            self.USEACORN=True
            print "Use Acorn to improve map set to: ", self.USEACORN
         if env_flag == 'USECPIRATE':
            self.USECPIRATE=True
            print "Use Cpirate to improve phases set to: ", self.USECPIRATE
         if env_flag == 'USEENSEM':
            self.USEENSEM=True
            print "Use Ensemble of top models in MR set to: ", self.USEENSEM
         if env_flag == 'CHECK':
            self.CHECK=True
            print "Check web connection set to: ", self.CHECK
         if env_flag == 'ENANT':
            self.ENANT=True
            print "Enantiomporph search set to: ", self.ENANT
         if env_flag == 'DUMMY':
            self.DUMMY=True
            print "DUMMY run set to: ", self.DUMMY
         if env_flag == 'PICKLE':
            self.PICKLE=True
            print "PICKLE run set to: ", self.PICKLE
         if env_flag == 'DBOUT':
            self.DBOUT=True
            print "DBOUT run set to: ", self.DBOUT
         if env_flag == 'DBVIEW':
            self.DBVIEW=True
            print "DBVIEW run set to: ", self.DBVIEW
         if env_flag == 'REFTWIN':
            self.REFTWIN=True
            print "Refmac TWIN option run set to: ", self.REFTWIN

      if flag.upper() == 'FALSE' or flag == '0': 
         if env_flag == 'DEBUG': 
            self.DEBUG=False
            os.environ["MRBUMP_DEBUG"]="False"
            print "Debug set to: ", self.DEBUG
         if env_flag == 'CLUSTER': 
            self.CLUSTER=False
            os.environ["MRBUMP_CLUSTER"]="False"
            print "Cluster set to: ", self.CLUSTER
         if env_flag == 'SCOPSEARCH':
            self.SCOP=False
            print "SCOP search set to: ", self.SCOP
         if env_flag == 'SSMSEARCH':
            self.SSM=False
            print "SSM search set to: ", self.SSM
         if env_flag == 'PQSSEARCH':
            self.PQS=False
            print "PQS search set to: ", self.PQS
         if env_flag == 'PISASEARCH':
            self.PISA=False
            print "PISA search set to: ", self.PISA
         if env_flag == 'FASTALOCAL':
            self.fasta_local=False
            print "Run local Fasta search set to: ", self.fasta_local
         if env_flag == 'RUNLOCAL':
            self.run_local=False
            print "Run local set to: ", self.run_local
         if env_flag == 'UPDATE':
            self.UPDATE=False
            print "Update database files set to: ", self.UPDATE
         if env_flag == 'DOFASTA':
            self.DOFASTA=False
            print "Do FASTA search set to: ", self.DOFASTA
         if env_flag == 'TRYALL':
            self.TRYALL=False
            print "Run in TRYALL mode set to: ", self.TRYALL
         if env_flag == 'CLEAN':
            self.CLEAN=False
            print "CLEAN flag set to: ", self.CLEAN
         if env_flag == 'HTMLOUT':
            self.HTMLOUT=False
            print "Run in HTMLOUT mode set to: ", self.HTMLOUT
         if env_flag == 'XMLOUT':
            self.XMLOUT=False
            print "Run in XMLOUT mode set to: ", self.XMLOUT
         if env_flag == 'MDLDPDBCLP':
            self.MDLDPDBCLP=False
            print "Generate PDBClip models set to: ", self.MDLDPDBCLP
         if env_flag == 'MDLPLYALA':
            self.MDLPLYALA=False
            print "Generate Polyalanine models set to: ", self.MDLPLYALA
         if env_flag == 'MDLMOLREP':
            self.MDLMOLREP=False
            print "Generate Molrep models set to: ", self.MDLMOLREP
         if env_flag == 'MDLCHAINSAW':
            self.MDLCHAINSAW=False
            print "Generate Chainsaw models set to: ", self.MDLCHAINSAW
         if env_flag == 'USEACORN':
            self.USEACORN=False
            print "Use Acorn to improve map set to: ", self.USEACORN
         if env_flag == 'USECPIRATE':
            self.USECPIRATE=False
            print "Use Cpirate to improve phases set to: ", self.USECPIRATE
         if env_flag == 'USEENSEM':
            self.USEENSEM=False
            print "Use Ensemble in MR set to: ", self.USEENSEM
         if env_flag == 'CHECK':
            self.CHECK=False
            print "Check web connection set to: ", self.CHECK
         if env_flag == 'ENANT':
            self.ENANT=False
            print "Enantiomporph search set to: ", self.ENANT
         if env_flag == 'DUMMY':
            self.DUMMY=False
            print "DUMMY run set to: ", self.DUMMY
         if env_flag == 'PICKLE':
            self.PICKLE=False
            print "PICKLE run set to: ", self.PICKLE
         if env_flag == 'DBOUT':
            self.DBOUT=False
            print "DBOUT run set to: ", self.DBOUT
         if env_flag == 'DBVIEW':
            self.DBVIEW=False
            print "DBVIEW run set to: ", self.DBVIEW
         if env_flag == 'REFTWIN':
            self.REFTWIN=False
            print "Refmac TWIN option run set to: ", self.REFTWIN

   def parse_LABIN(self, data_line):
      """ A function to parse the LABIN keyword. """

      labels=[]

      # Check that the correct number of arguments are given
      if data_line.count('=') != 3:
         self.keyword_usage('labin', 'Incorrect number of arguments for LABIN', 1)

      # Split the line into its separate parts
      line=string.split(data_line,'=')

      # Remove the 'LABIN' word from the start of the line
      labin_part=string.split(line[0])
      del labin_part[0]

      # Remove the first entry and re-enter it without 'LABIN'
      del line[0]
      line.insert(0,labin_part[0])

      # Now split the line again
      for i in line:
        t=string.split(i)
        for j in t:
           labels.append(j)

      # Assign the column labels from the list (0=1, 2=3 4=5)
      self.col_labels[labels[0].upper()]=labels[1]
      self.col_labels[labels[2].upper()]=labels[3]
      self.col_labels[labels[4].upper()]=labels[5]

      # Check that all of the necessary column labels are given
      print ""
      for i in 'F', 'SIGF' , 'FREER_FLAG':
         if i not in self.col_labels.keys(): 
            self.keyword_usage('labin', "Column label %s not given" % i, 2)
         print "column %s labelled as %s" % (i, self.col_labels[i]) 
      print ""

   def verify_NUM(self):
      """ A function to chekc that the NUM_ variables adhere to the constraints. """

      # if we are not doing Ensemble, set the number in the ensemble to 0
      if self.USEENSEM == False:
         self.ENSEMNUM = 0

      # Verify that ENSEMNUM is less than MRNUM
      if self.ENSEMNUM > self.MRNUM and "PHASER" in self.MR_PROGRAM_LIST:
         sys.stdout.write("Input Warning: Number of search models to be used in Phaser ensemble (%d) is more than the\n" % self.ENSEMNUM
                        + "               number of models to be prepared for MR (%d). Reseting the number in the ensemble\n" % self.MRNUM
                        + "               to the number of search models being prepared.\n")
         sys.stdout.write("\n")

         self.ENSEMNUM = self.MRNUM
         sys.stdout.write("Number of models to use in Phaser ensemble: %d\n" % self.ENSEMNUM)
         sys.stdout.write("\n")

   def keyword_usage(self, keyword, error_string, error_code):
      """ A function to echo the correct usage of the keywords to the stdout. """

      self.setErrorString(error_string)
 
      print "Keyword error:"

      # LABIN errors:
      if keyword == 'labin':
         print "LABIN error: %s" % self.error_string
         print "Keyword LABIN must be provided and must specify 3 column labels: F, SIGF and FreeR_flag"
         print "  e.g. LABIn F=FP SIGF=SIGFP FreeR_flag=FREE"
     
      # JOBID errors
      if keyword == 'jobid':
         print "JOBID error: %s" % self.error_string
         print "Keyword JOBID must be provided"
         print "  e.g. JOBID Test_job1"
     
      # ROOTDIR errors
      if keyword == 'root_dir':
         print "ROOTDIR error: %s" % self.error_string
     
      # RESULTS HTML errors
      if keyword == 'reshtml':
         print "RESHTML error: %s" % self.error_string
     
      # IGNORE PDB id errors
      if keyword == 'ignore':
         print "IGNORE error: %s" % self.error_string
         print "Keyword IGNORE must specify PDB codes"
         print "  e.g. IGNORE 1smw 1smm 1smu 1ahc ......"

      # INCLUDE PDB id errors
      if keyword == 'include':
         print "INCLUDE error: %s" % self.error_string
         print "Keyword INCLUDE must specify chain codes. This is a four character PDB code\n" \
               + "followed by an underscore and a chain identifier letter."
         print "  e.g. INCLUDE 1smw_A 1smm_B 1smu_A 1ahc_C ......"

      # NMASU errors
      if keyword == 'NMASU':
         print "%s error: %s" % (keyword, self.error_string)
         print "  e.g. %s 4" % keyword
    
      # PACK errors
      if keyword == 'PACK':
         print "%s error: %s" % (keyword, self.error_string)
         print "  e.g. %s 4" % keyword
    
      # NCYC errors
      if keyword == 'NCYC':
         print "%s error: %s" % (keyword, self.error_string)
         print "  e.g. %s 10" % keyword
    
      # PIRCYC errors
      if keyword == 'PIRCYC':
         print "%s error: %s" % (keyword, self.error_string)
         print "  e.g. %s 10" % keyword
    
      # NUM_values errors
      if keyword == 'NUM_values':
         print "'NUM' variables error: %s" % self.error_string

      if keyword == 'MRNUM' or keyword == 'ENSEMNUM':
         print "%s error: %s" % (keyword, self.error_string)
         print "  e.g. %s 20" % keyword
    
      # EVALUE errors
      if keyword == 'E_value':
         print "EVALUE error: %s" % self.error_string
         print "  e.g. EVALUE 0.02"

      # ACORNRES errors
      if keyword == 'ACORNRES':
         print "ACORNRES error: %s" % self.error_string
         print "  e.g. ACORNRES 1.7"

      # Debug errors
      if keyword == 'DEBUG' or keyword == 'CLUSTER' or keyword == 'SCOPSEARCH' or keyword == 'SSMSEARCH' \
            or keyword == 'PQSSEARCH' or keyword == 'PWALIGN' or keyword == 'FASTALOCAL' \
            or keyword == 'RUNLOCAL' or keyword == 'UPDATE' \
            or keyword == 'MDLCHAINSAW' or keyword == 'MDLMOLREP' \
            or keyword == 'MDLPLYALA' or keyword == 'MDLDPDBCLP' or keyword == 'TRYALL' \
            or keyword == 'HTMLOUT' or keyword == 'XMLOUT' or keyword == 'USEACORN' or keyword == 'USECPIRATE' \
            or keyword == 'ENANT' or keyword == 'USEENSEM' or keyword == 'PISASEARCH' \
            or keyword == 'DUMMY' or keyword == 'PICKLE' or keyword == 'DBOUT' \
            or keyword == 'DBVIEW' or keyword == 'CLEAN' or keyword == "CHECK":  
         print "%s error: %s" % (keyword, self.error_string)
         print "  e.g. %s True" % keyword

      # MRPROGRAM errors
      if keyword == 'MRPROGRAMS':
         print "MRPROGRAMS error: %s" % self.error_string
         print "Keyword MRPROGRAMS must be 'MOLREP' or 'PHASER' or both"
         print "  e.g. MRPROGRAMS PHASER MOLREP"

      # MAPROGRAM errors
      if keyword == 'MAPROGRAM':
         print "MAPROGRAM error: %s" % self.error_string
         print "Keyword MAPROGRAM must be 'MAFFT', 'CLUSTALW', 'CLUSTALW2', 'T_COFFEE' or 'PROBCONS'"
         print "  e.g. MAPROGRAM Mafft"

      # QTYPE errors
      if keyword == 'QTYPE':
         print "QTYPE error: %s" % self.error_string
         print "Keyword QTYPE must be 'SGE' or 'PBS'"
         print "  e.g. QTYPE SGE"

      # FASTA search error
      if keyword == 'DOFASTA':
         print "DOFASTA error: %s" % self.error_string

      sys.exit(error_code)

   def get_FILENAME(self, data_line, flags=[]):
      """ get a filename from a keyword line (accounting for spaces). """

      name=[]
      line=string.split(data_line)
      del line[0]
      for i in range(len(line)):
         if line[i].upper() not in flags:
            name.append(line[i])
         else:
            break

      filename=string.joinfields(name, " ")
     
      return filename

   def is_an_integer(self, s):
      """ A quick function to check if a string is an integer. """
   
      try: int(s)
      except ValueError: return False
      else: return True   

   def is_a_float(self, s):
      """ A quick function to check if a string is a float. """
   
      try: float(s)
      except ValueError: return False
      else: return True   


class FixedPDB:
   """ An object to hold details for a Fixed PDB file in the MR search. """

   def __init__(self):
      self.filename=""
      self.identity=0.0
      self.mol_weight=0.0


class LocalPDB:
   """ An object to hold the details of a local file specified by the user for inclusionn
       in the molecular replacement step. """
   
   def __init__(self):
      self.filename=""
      self.filestring=""
      self.chainID=""
      self.rms=0.0
      self.number=0


if __name__ == '__main__':

   k=Keywords()
   k.parse_LABIN("LABIn FP=FP SIGF=SIGFP freer_flag=FREE")
