#! /usr/bin/env python
#
#     Copyright (C) 2005 Ronan Keegan
#
#     This code is distributed under the terms and conditions of the
#     CCP4 Program Suite Licence Agreement as a CCP4 Application.
#     A copy of the CCP4 licence can be obtained by writing to the
#     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
#
#
#
# Wrapper for clustalw the sequence multiple alignment program
# Ronan Keegan 13/9/05

import os, sys, string
import Scan_path


class Clustalw:
   """ A class to call clustalw to do a multiple alignment of the target and template sequences. """
 
   def __init__(self):

      chk=Scan_path.Check_PATH()
      if chk.find_exec("clustalw2"):
         self.clustalw_EXE=os.path.join(chk.exe_directory, "clustalw2")
      elif chk.find_exec("clustalw"):
         self.clustalw_EXE=os.path.join(chk.exe_directory, "clustalw")
      else:
         sys.stdout.write("Clustalw Error: clustalw and clustalw2 executables not found on system.\n")
         sys.stdout.write("\n")
         sys.exit()

      self.input_fasta_file = ''
      self.output_pir_file = ''
      self.log_file = ''
      try:
         self.debug=eval(os.environ['MRBUMP_DEBUG'])
      except:
         self.debug=False

   def setDEBUG(self, flag):
      self.debug=flag  

   def setInputFastaFile(self, file_name):
      self.input_fasta_file = file_name

   def setOutputPIRFile(self, file_name):
      self.output_pir_file = file_name

   def setLogFile(self, file_name):
      self.log_file = file_name

   def Clustalw(self, input_fasta_file, output_pir_file, logfile):
      """ A function to do a system call to Clustalw."""

      self.setInputFastaFile(input_fasta_file)
      self.setOutputPIRFile(output_pir_file)
      self.setLogFile(logfile)

      if self.debug:
         print ""
         print "======================"
         print "Clustalw command line:"
         print "======================"
         print self.clustalw_EXE + ' -align -outorder=input -output=pir -infile=' + self.input_fasta_file + \
                ' -outfile=' + self.output_pir_file
         print ""

      # Open up a pipe to the clustalw program
      c,oe=os.popen4(self.clustalw_EXE + ' -align -outorder=input -output=pir -infile=' + self.input_fasta_file + \
             ' -outfile=' + self.output_pir_file)
      c.close()
   
      # Capture any error from the alignment job and print it to screen if debug mode is on
      log=open(self.log_file, "w")
      line=oe.readline()
      while line:
         if self.debug:
            log.write(line)
         line=oe.readline()
      oe.close()
      log.close()
      
   def convert_to_fasta(self):
  
      infile=open(self.output_pir_file,'r')
      
      line=infile.readline()
      while line: 
         print line

         line=infile.readline()
   
   def get_raw_target_len(self):
      ''' A function to get the length of the target sequence before it is processed by the alignment program. '''

      seq = ''
 
      file = open(self.input_fasta_file, 'r')

      line = file.readline()
      name=string.strip(line).replace('>','') 
      line = file.readline()
      while line:
         if '>' in line:
            break
         else:
           seq = seq + string.strip(line)
         line = file.readline()
    
      return len(seq)
         
  
if __name__ == '__main__':

   clustalw=Clustalw_multiple()
   clustalw.Clustalw('input/1vrd_all.fasta', 'out_1vrd.pir', 'clustalw.log')
   clustalw.convert_to_fasta()



